import concurrent.futures
import hashlib
from typing import Any, Dict, List

import boto3
import requests
from botocore.client import BaseClient, Config
from botocore.exceptions import ClientError

from paka.kube_resources.model_group.manifest import Manifest
from paka.kube_resources.model_group.supported_models import SUPPORTED_MODELS
from paka.logger import logger
from paka.utils import read_current_cluster_data, to_yaml

MODEL_PATH_PREFIX = "models"


def delete_s3_file(bucket_name: str, s3_file_name: str) -> None:
    s3 = boto3.client("s3")
    if s3_file_exists(bucket_name, s3_file_name):
        s3.delete_object(Bucket=bucket_name, Key=s3_file_name)
        logger.info(f"{s3_file_name} deleted.")
    else:
        logger.info(f"{s3_file_name} not found.")


def s3_file_exists(bucket_name: str, s3_file_name: str) -> bool:
    s3 = boto3.client("s3")
    try:
        s3.head_object(Bucket=bucket_name, Key=s3_file_name)
        return True
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        else:
            raise  # some other error occurred


def s3_file_prefix_exists(bucket_name: str, s3_file_name: str) -> bool:
    s3 = boto3.resource("s3")
    bucket = s3.Bucket(bucket_name)
    return any(bucket.objects.filter(Prefix=s3_file_name))


def save_string_to_s3(bucket_name: str, file_name: str, data: str) -> None:
    s3 = boto3.resource("s3")
    s3.Object(bucket_name, file_name).put(Body=data)


def upload_part(
    s3: BaseClient,
    bucket_name: str,
    s3_file_name: str,
    upload_id: str,
    part_number: int,
    chunk: bytes,
) -> Dict[str, Any]:
    part = s3.upload_part(
        Body=chunk,
        Bucket=bucket_name,
        Key=s3_file_name,
        UploadId=upload_id,
        PartNumber=part_number,
    )
    return {"PartNumber": part_number, "ETag": part["ETag"]}


def download_file_to_s3(
    url: str,
    bucket_name: str,
    s3_file_name: str,
    chunk_size: int = 5 * 1024 * 1024,
    max_parallel_uploads: int = 20,
) -> str:
    """
    Downloads a file from a URL and uploads it to an S3 bucket in chunks.

    This function downloads a file from the specified URL and uploads it to the specified S3 bucket.
    The file is uploaded in chunks of the specified size. The function uses a ThreadPoolExecutor to
    upload the chunks in parallel, with the specified maximum number of parallel uploads.

    The function calculates the SHA256 hash of the file while it is being downloaded and uploaded.
    If the upload is successful, the function returns the SHA256 hash.

    If an error occurs during the download or upload, the function logs the error and raises an exception.
    If an error occurs during the upload, the function aborts the multipart upload.

    Args:
        url (str): The URL to download the file from.
        bucket_name (str): The name of the S3 bucket to upload the file to.
        s3_file_name (str): The name to give to the file in the S3 bucket.
        chunk_size (int, optional): The size of the chunks to upload. Defaults to 5 * 1024 * 1024.
        max_parallel_uploads (int, optional): The maximum number of parallel uploads. Defaults to 20.

    Raises:
        Exception: If an error occurs during the download or upload.

    Returns:
        str: The SHA256 hash of the file.
    """

    s3 = boto3.client("s3", config=Config(signature_version="s3v4"))
    upload_id = None
    upload_completed = False

    try:
        with requests.get(url, stream=True) as response:
            if response.status_code == 200:
                sha256 = hashlib.sha256()
                total_size = int(response.headers.get("content-length", 0))
                processed_size = 0

                upload = s3.create_multipart_upload(
                    Bucket=bucket_name, Key=s3_file_name
                )
                upload_id = upload["UploadId"]
                parts = []

                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=max_parallel_uploads
                ) as executor:
                    futures: List[concurrent.futures.Future] = []
                    part_number = 1

                    for chunk in response.iter_content(chunk_size=chunk_size):
                        sha256.update(chunk)
                        while len(futures) >= max_parallel_uploads:
                            # Wait for one of the uploads to complete
                            done, _ = concurrent.futures.wait(
                                futures, return_when=concurrent.futures.FIRST_COMPLETED
                            )
                            for future in done:
                                parts.append(future.result())
                                futures.remove(future)

                        # Submit new chunk for upload
                        future = executor.submit(
                            upload_part,
                            s3,
                            bucket_name,
                            s3_file_name,
                            upload_id,
                            part_number,
                            chunk,
                        )
                        futures.append(future)
                        part_number += 1
                        processed_size += len(chunk)
                        progress = (processed_size / total_size) * 100
                        print(f"Progress: {progress:.2f}%", end="\r")

                    # Wait for all remaining uploads to complete
                    for future in concurrent.futures.as_completed(futures):
                        parts.append(future.result())

                parts.sort(key=lambda part: part["PartNumber"])
                s3.complete_multipart_upload(
                    Bucket=bucket_name,
                    Key=s3_file_name,
                    UploadId=upload_id,
                    MultipartUpload={"Parts": parts},
                )

                upload_completed = True
                logger.info(f"File uploaded to S3: {s3_file_name}")
                sha256_value = sha256.hexdigest()
                logger.info(f"SHA256 hash of the file: {sha256_value}")
                return sha256_value
            else:
                error_message = f"Unable to download the file. HTTP Status Code: {response.status_code}"
                logger.error(error_message)
                raise Exception(error_message)

    except Exception as e:
        logger.error(f"An error occurred: {str(e)}")
        raise

    finally:
        if upload_id is not None and not upload_completed:
            # If an error occurred and upload was not completed
            s3.abort_multipart_upload(
                Bucket=bucket_name, Key=s3_file_name, UploadId=upload_id
            )


def download_model(name: str) -> None:
    """
    Downloads a machine learning model from a URL and uploads it to an S3 bucket.

    This function checks if the model is supported and if it already exists in the S3 bucket.
    If the model is supported and does not exist, the function downloads the model from the URL
    and uploads it to the S3 bucket. The function also saves a manifest file in the S3 bucket
    with metadata about the model.

    If the SHA256 hash of the downloaded file does not match the expected hash, the function
    deletes the file from the S3 bucket and raises an exception.

    Args:
        name (str): The name of the model to download.

    Raises:
        Exception: If the model is not supported or if the SHA256 hash of the downloaded file does not match.

    Returns:
        None
    """
    if name not in SUPPORTED_MODELS:
        logger.error(
            f"Model {name} is not supported."
            f"Available models are: {', '.join(SUPPORTED_MODELS.keys())}"
        )
        raise Exception(f"Model {name} is not supported.")

    model = SUPPORTED_MODELS[name]

    logger.info(f"Downloading model from {model.url}...")
    # Get the model name from the URL
    model_file_name = model.url.split("/")[-1]
    model_path = f"{MODEL_PATH_PREFIX}/{name}"

    full_model_file_path = f"{model_path}/{model_file_name}"
    bucket = read_current_cluster_data("bucket")

    if s3_file_prefix_exists(bucket, f"{model_path}/"):
        logger.info(f"Model {name} already exists.")
        return

    sha256 = download_file_to_s3(model.url, bucket, full_model_file_path)
    if sha256 != model.sha256:
        logger.error(f"SHA256 hash of the downloaded file does not match.")
        # Delete the file
        delete_s3_file(bucket, full_model_file_path)
        raise Exception(f"SHA256 hash of the downloaded file does not match.")

    # Save model manifest
    manifest = Manifest(
        name=name,
        sha256=model.sha256,
        url=model.url,
        type="gguf",  # TODO: hard-coded for now
        file=model_file_name,
    )

    manifest_yaml = to_yaml(manifest.model_dump(exclude_none=True))
    save_string_to_s3(bucket, f"{model_path}/manifest.yaml", manifest_yaml)

    logger.info(f"Model {name} downloaded successfully.")


def get_model_file_name(model_name: str) -> str:
    """
    Returns the file name of a machine learning model.

    This function checks if the model is supported and if so, returns the file name of the model.
    The file name is extracted from the model's URL.

    Args:
        model_name (str): The name of the model.

    Raises:
        Exception: If the model is not supported.

    Returns:
        str: The file name of the model.
    """
    if model_name not in SUPPORTED_MODELS:
        logger.error(
            f"Model {model_name} is not supported."
            f"Available models are: {', '.join(SUPPORTED_MODELS.keys())}"
        )
        raise Exception(f"Model {model_name} is not supported.")

    model = SUPPORTED_MODELS[model_name]

    # Get the model name from the URL
    model_name = model.url.split("/")[-1]
    return model_name



import typer

from paka.cli.utils import build as build_image

build_app = typer.Typer()


@build_app.callback(invoke_without_command=True)
def build(
    source_dir: str = typer.Argument(
        ...,
        help="Source directory of the application.",
    ),
    image_name: str = typer.Option(
        "",
        "--image-name",
        help="Provide a custom name for the Docker image. If omitted, "
        "the base name of the source code directory will be used as the image name.",
    ),
) -> None:
    """
    Build a Docker image from the application in the specified source directory.
    The source directory must contain a Procfile and a .cnignore file. The Procfile
    defines the commands to run for the application. The .cnignore file defines the
    files and directories to exclude from the image. Once the image is built,
    it will be pushed to the container repository of the current cluster.

    A Dockerfile is NOT required. The image will be built using Cloud Native Buildpacks.
    In cluster build is not supported yet. User machine must have Docker installed.
    """
    build_image(source_dir, image_name)



import re
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, field_validator, model_validator
from ruamel.yaml import YAML

from paka.utils import to_yaml


def validate_size(v: str, error_message: str = "Invalid size format") -> str:
    """
    Validates the format of the size field.

    Args:
        v (str): The value of the size field.
        error_message (str, optional): The error message to raise if the format of the input value is invalid. Defaults to "Invalid size format".

    Returns:
        str: The input value if validation is successful.

    Raises:
        ValueError: If the format of the input value is invalid. The error message is specified by the `error_message` parameter.
    """
    if not re.match(r"^\d+(Mi|Gi)$", v):
        raise ValueError(error_message)
    return v


class ResourceRequest(BaseModel):
    """
    Represents the resource request for a container.

    Attributes:
        cpu (str): The amount of CPU to request.
        memory (str): The amount of memory to request.
        gpu (Optional[int]): The number of GPUs to request. Defaults to None.
    """

    cpu: str
    memory: str
    gpu: Optional[int] = None

    @field_validator("cpu", mode="before")
    def validate_cpu(cls, v: str) -> str:
        """
        Validates the format of the cpu field.

        Args:
            v (str): The value of the cpu field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        if not re.match(r"^\d+(m)?$", v):
            raise ValueError("Invalid CPU format")
        return v

    @field_validator("memory", mode="before")
    def validate_memory(cls, v: str) -> str:
        """
        Validates the format of the memory field.

        Args:
            v (str): The value of the memory field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid memory format")

    @field_validator("gpu")
    def validate_gpu(cls, v: Optional[int]) -> Optional[int]:
        """
        Validates the value of the gpu field.

        Args:
            v (Optional[int]): The value of the gpu field.

        Returns:
            Optional[int]: The input value if validation is successful.

        Raises:
            ValueError: If the value is less than 0.
        """
        if v is not None and v < 0:
            raise ValueError("GPU count cannot be less than 0")
        return v


class AwsGpuNode(BaseModel):
    """
    Represents a configuration for an AWS GPU node.

    Attributes:
        diskSize (int): The size of the disk for the GPU node in GB.
    """

    diskSize: int


class GcpGpuNode(BaseModel):
    """
    Represents a Google Cloud Platform GPU node.

    Attributes:
        imageType (str): The type of image used for the GPU node.
        acceleratorType (str): The type of accelerator used for the GPU node.
        acceleratorCount (int): The number of accelerators attached to the GPU node.
        diskType (str): The type of disk used for the GPU node.
        diskSize (int): The size of the disk attached to the GPU node in GB.
    """

    imageType: str
    acceleratorType: str
    acceleratorCount: int
    diskType: str
    diskSize: int


class CloudNode(BaseModel):
    """
    Represents a node in the cloud cluster.

    Attributes:
        nodeType (str): The type of the node.
        diskSize (int): The size of the disk attached to the node in GB.
        awsGpu (Optional[AwsGpuNode]): The AWS GPU node configuration, if applicable.
        gcpGpu (Optional[GcpGpuNode]): The GCP GPU node configuration, if applicable.
    """

    nodeType: str
    diskSize: int = 20
    awsGpu: Optional[AwsGpuNode] = None
    gcpGpu: Optional[GcpGpuNode] = None

    @model_validator(mode="before")
    def validate_gpu(
        cls, values: Dict[str, Union[AwsGpuNode, GcpGpuNode]]
    ) -> Dict[str, Union[AwsGpuNode, GcpGpuNode]]:
        if values.get("awsGpu") and values.get("gcpGpu"):
            raise ValueError("At most one of awsGpu or gcpGpu can exist")
        return values


class ModelGroup(BaseModel):
    """
    Represents a group of VMs that serve the inference for a specific type of model.

    Attributes:
        name (str): The name of the model group.
        minInstances (int): The minimum number of instances to provision.
        maxInstances (int): The maximum number of instances to provision.
    """

    name: str
    minInstances: int
    maxInstances: int

    @model_validator(mode="before")
    def check_instances_num(cls, values: Dict[str, int]) -> Dict[str, int]:
        min_instances, max_instances = values.get("minInstances"), values.get(
            "maxInstances"
        )
        if min_instances and max_instances and max_instances < min_instances:
            raise ValueError(
                "maxInstances must be greater than or equal to minInstances"
            )
        return values

    @field_validator("minInstances", mode="before")
    def validate_min_instances(cls, v: int) -> int:
        """
        Validates the value of the minInstances field. Spinning up 1 model group instance is heavy.
        No scaling down to 0 for now.

        Args:
            v (int): The value of the minInstances field.

        Returns:
            int: The input value if validation is successful.

        Raises:
            ValueError: If the value of the input value is invalid.
        """
        if v <= 0:
            raise ValueError("minInstances must be greater than 0")
        return v


class Trigger(BaseModel):
    type: str
    metadata: Dict[str, str]


class CloudModelGroup(ModelGroup, CloudNode):
    """
    Represents a group of cloud models.

    This class inherits from both the `ModelGroup` and `CloudNode` classes.

    Attributes:
        resourceRequest (Optional[ResourceRequest]): The resource request for the model group, specifying the amount of CPU and memory to request.

    Inherited Attributes:
        name (str): The name of the model group.
        minInstances (int): The minimum number of instances to provision for the model group.
        maxInstances (int): The maximum number of instances to provision for the model group.
        nodeType (str): The type of the node.
    """

    # TODO: make required for HPA to work
    resourceRequest: Optional[ResourceRequest] = None

    autoScaleTriggers: Optional[List[Trigger]] = None
    """
    Triggers for autoscaling the model group. Trigger are not strongly typed, so we use a list of dictionaries to represent them.
    For example:

    autoScaleTriggers=[
        # CPU trigger example
        Trigger(
            type="cpu",
            metadata={
                "type": "Utilization",
                "value": "70"
            }
        ),
        # Prometheus trigger example
        Trigger(
            type="prometheus",
            metadata={
                "serverAddress": "http://prometheus-operated.default.svc.cluster.local",
                "metricName": "http_requests_total",
                "threshold": "100",
                "query": "sum(rate(http_requests_total{job=\"example-job\"}[2m]))"
            }
        )
    ]
    """


class ClusterConfig(BaseModel):
    """
    Represents the configuration for a cluster.

    Attributes:
        name (str): The name of the cluster.
        region (str): The default region for the cluster.
        namespace (str, optional): The namespace in which the cluster is deployed. Defaults to 'default'.
        nodeType (str): The type of nodes in the cluster.
        minNodes (int): The minimum number of nodes in the cluster.
        maxNodes (int): The maximum number of nodes in the cluster.
        logRetentionDays (int, optional): The number of days to retain log entries. Defaults to 14.
    """

    name: str
    region: str

    namespace: str = "default"

    nodeType: str
    minNodes: int
    maxNodes: int

    logRetentionDays: int = 14


class CloudVectorStore(CloudNode):
    """
    Represents a cloud vector store.

    Attributes:
        replicas (int): The number of replicas for the vector store. Defaults to 1.
        storage_size (str): The size of the storage of one node for the vector store. Defaults to "10Gi".
        resourceRequest (Optional[ResourceRequest]): The resource request for the vector store, specifying the amount of CPU and memory to request.

    Inherited Attributes:
        nodeType (str): The type of the node.

    Methods:
        validate_storage_size: Validates the format of the storage_size field.
    """

    replicas: int = 1
    storage_size: str = "10Gi"
    resourceRequest: Optional[ResourceRequest] = None

    @field_validator("storage_size", mode="before")
    def validate_storage_size(cls, v: str) -> str:
        """
        Validates the format of the storage_size field.

        Args:
            v (str): The value of the storage_size field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid storage size format")

    @field_validator("replicas", mode="before")
    def validate_replicas(cls, v: int) -> int:
        """
        Validates the value of the replicas field.

        Args:
            v (int): The value of the replicas field.

        Returns:
            int: The input value if validation is successful.

        Raises:
            ValueError: If the value of the input value is invalid.
        """
        if v <= 0:
            raise ValueError("replicas must be greater than 0")
        return v


class Job(BaseModel):
    """
    Represents a job cluster configuration.

    Attributes:
        broker_storage_size (str): The size of the storage for the broker. Defaults to "10Gi".
    """

    enabled: bool = False

    broker_storage_size: str = "10Gi"

    @field_validator("broker_storage_size", mode="before")
    def validate_broker_storage_size(cls, v: str) -> str:
        """
        Validates the format of the broker_storage_size field.

        Args:
            v (str): The value of the storage_size field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid storage size format")


class Prometheus(BaseModel):
    enabled: bool = False

    storage_size: str = "10Gi"
    grafana: bool = False
    alertmanager: bool = False
    kube_api_server: bool = False
    kubelet: bool = False
    kube_controller_manager: bool = False
    core_dns: bool = False
    kube_etcd: bool = False
    kube_scheduler: bool = False
    kube_proxy: bool = False
    kube_state_metrics: bool = False
    node_exporter: bool = False
    thanos_ruler: bool = False

    @field_validator("storage_size", mode="before")
    def validate_storage_size(cls, v: str) -> str:
        """
        Validates the format of the storage_size field.

        Args:
            v (str): The value of the storage_size field.

        Returns:
            str: The input value if validation is successful.

        Raises:
            ValueError: If the format of the input value is invalid.
        """
        return validate_size(v, "Invalid storage size format")


class Tracing(BaseModel):
    """Represents the configuration for tracing."""

    enabled: bool = False

    autoScalingEnabled: bool = False

    zipkinHelmSettings: Optional[Dict[str, Any]] = None
    """https://github.com/openzipkin/zipkin-helm"""


class CloudConfig(BaseModel):
    """
    Represents the configuration for the cloud environment.

    Attributes:
        cluster (ClusterConfig): The configuration for the Kubernetes cluster.
        modelGroups (List[CloudModelGroup], optional): The list of model groups
            to be deployed in the cloud. If None, no model groups will be deployed. Defaults to None.
        vectorStore (CloudVectorStore, optional): The configuration
            for the vector store in the cloud. If None, the vector store will
            not be provisioned. Defaults to None.
        job (Job, optional): The configuration for the job broker. If None, the
            job broker will not be provisioned. Defaults to None.
        prometheus (Prometheus, optional): The configuration for Prometheus.
            If None, Prometheus will not be provisioned. Defaults to None.
    """

    cluster: ClusterConfig
    modelGroups: Optional[List[CloudModelGroup]] = None
    vectorStore: Optional[CloudVectorStore] = None
    job: Optional[Job] = None

    prometheus: Optional[Prometheus] = None
    tracing: Optional[Tracing] = None

    @field_validator("modelGroups", mode="before")
    def check_model_group(cls, v: List[Any]) -> List[Any]:
        if v is None or len(v) == 0:
            return v
        # Check if there are duplicated model group names
        model_group_names = [dict(group)["name"] for group in v]
        if len(set(model_group_names)) != len(model_group_names):
            raise ValueError("Duplicate model group names are not allowed")
        return v


class Config(BaseModel):
    """
    Configuration class for managing cloud cluster settings.

    Attributes:
        aws (Optional[CloudConfig]): AWS cloud configuration.
        gcp (Optional[CloudConfig]): GCP cloud configuration.

    Methods:
        check_one_field: Validates that exactly one cloud configuration (aws or gcp) is provided.
    """

    aws: Optional[CloudConfig] = None
    gcp: Optional[CloudConfig] = None

    @model_validator(mode="before")
    def check_one_field(cls, values: Dict[str, CloudConfig]) -> Dict[str, CloudConfig]:
        """
        Validates that exactly one cloud configuration is provided.

        Args:
            values (Dict[str, CloudConfig]): Dictionary of field values, where keys are the cloud providers (aws, gcp) and values are the corresponding cloud configurations.

        Returns:
            Dict[str, CloudConfig]: The input values if validation is successful.

        Raises:
            ValueError: If more or less than one cloud configuration is provided.
        """
        non_none_fields = sum(value is not None for value in values.values())
        if non_none_fields != 1:
            raise ValueError("Exactly one cloud configuration must be provided")

        return values


def generate_yaml(config: Config) -> str:
    """
    Generate a YAML string representation of the given config object.

    Args:
        config (Config): The config object to generate YAML from.

    Returns:
        str: The YAML string representation of the config object.
    """
    return to_yaml(config.model_dump(exclude_none=True))


def parse_yaml(yaml_str: str) -> Config:
    """
    Parse a YAML string and return a Config object.

    Args:
        yaml_str (str): The YAML string to parse.

    Returns:
        Config: The parsed Config object.
    """
    yaml = YAML()
    data = yaml.load(yaml_str)
    return Config(**data)



from typing import Any, List, Optional

from kubernetes import client

from paka.config import CloudConfig, CloudModelGroup, Config
from paka.constants import ACCESS_ALL_SA
from paka.k8s import CustomResource, apply_resource, try_load_kubeconfig
from paka.kube_resources.model_group.model import MODEL_PATH_PREFIX, download_model
from paka.utils import kubify_name, read_cluster_data

# `latest` will be stale because of the `IfNotPresent` policy
# We hardcode the image here for now, we can make it configurable later
LLAMA_CPP_PYTHON_IMAGE = "ghcr.io/abetlen/llama-cpp-python:latest"

LLAMA_CPP_PYTHON_CUDA = "jijunleng/llama-cpp-python-cuda:latest"

try_load_kubeconfig()


def init_aws(config: CloudConfig, model_group: CloudModelGroup) -> client.V1Container:
    """
    Initializes an AWS container for downloading a model from S3.

    Args:
        config (CloudConfig): The cloud configuration.
        model_group (CloudModelGroup): The cloud model group.

    Returns:
        client.V1Container: The initialized AWS container.
    """
    bucket = read_cluster_data(config.cluster.name, "bucket")

    return client.V1Container(
        name="init-s3-model-download",
        image="amazon/aws-cli",
        command=[
            "aws",
            "s3",
            "cp",
            f"s3://{bucket}/{MODEL_PATH_PREFIX}/{model_group.name}/",
            "/data/",
            "--recursive",
        ],
        volume_mounts=[
            client.V1VolumeMount(
                name="model-data",
                mount_path="/data",
            )
        ],
    )


def init_parse_manifest() -> client.V1Container:
    """
    Initializes a container for creating a symbolic link to the model file.

    Returns:
        client.V1Container: The initialized container for handling the manifest.
    """
    return client.V1Container(
        name="init-parse-manifest",
        image="busybox",
        command=[
            "sh",
            "-c",
            "model_type=$(cat /data/manifest.yaml | grep 'type' | cut -d ':' -f2 | tr -d ' '); "
            'if [ "$model_type" != "gguf" ]; then echo \'Invalid model type\' && exit 1; fi; '
            "model_file=$(cat /data/manifest.yaml | grep 'file' | cut -d ':' -f2 | tr -d ' '); "
            "ln -s /data/${model_file} /data/my_model.gguf",
        ],
        volume_mounts=[
            client.V1VolumeMount(
                name="model-data",
                mount_path="/data",
            )
        ],
    )


def create_pod(
    namespace: str,
    config: Config,
    model_group: CloudModelGroup,
    runtime_image: str,
    port: int,
) -> client.V1Pod:
    """
    Creates a Kubernetes Pod for a machine learning model group.

    This function creates a Kubernetes Pod with the specified configuration. The Pod runs a container
    with the specified runtime image and exposes the specified port. The container runs a machine learning
    model from the model group.

    Args:
        namespace (str): The namespace to create the Pod in.
        config (Config): The configuration for the Pod.
        model_group (CloudModelGroup): The model group to run in the Pod.
        runtime_image (str): The runtime image for the container.
        port (int): The port to expose on the container.

    Raises:
        ValueError: If the AWS configuration is not provided.

    Returns:
        client.V1Pod: The created Pod.
    """
    if config.aws is None:
        raise ValueError("Only AWS is supported at this time")

    container_args = {
        "name": f"{kubify_name(model_group.name)}",
        "image": runtime_image,
        "volume_mounts": [
            client.V1VolumeMount(
                name="model-data",
                mount_path="/data",
            )
        ],
        "env": [
            client.V1EnvVar(
                name="N_GPU_LAYERS",
                # -1 means all layers are GPU layers, 0 means no GPU layers
                value=("-1" if model_group.awsGpu else "0"),
            ),
            client.V1EnvVar(
                name="MODEL",
                value=f"/data/my_model.gguf",
            ),
            client.V1EnvVar(
                name="PORT",
                value=str(port),
            ),
        ],
        "readiness_probe": client.V1Probe(
            http_get=client.V1HTTPGetAction(
                path="/v1/models",
                port=port,
            ),
            initial_delay_seconds=60,
            period_seconds=5,
            timeout_seconds=30,
            success_threshold=1,
            failure_threshold=5,
        ),
        "liveness_probe": client.V1Probe(
            http_get=client.V1HTTPGetAction(
                path="/v1/models",
                port=port,
            ),
            initial_delay_seconds=240,
            period_seconds=30,
            timeout_seconds=30,
            success_threshold=1,
            failure_threshold=5,
        ),
    }

    if model_group.resourceRequest:
        container_args["resources"] = client.V1ResourceRequirements(
            requests={
                "cpu": model_group.resourceRequest.cpu,
                "memory": model_group.resourceRequest.memory,
            },
        )

    if model_group.awsGpu:
        if "resources" not in container_args:
            container_args["resources"] = client.V1ResourceRequirements()
        if container_args["resources"].limits is None:
            container_args["resources"].limits = {}
        gpu_count = 1
        if model_group.resourceRequest and model_group.resourceRequest.gpu:
            gpu_count = model_group.resourceRequest.gpu
        # Ah, we only support nvidia GPUs for now
        container_args["resources"].limits["nvidia.com/gpu"] = gpu_count

    return client.V1Pod(
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
            labels={
                "app": "model-group",
                "model": model_group.name,
            },
        ),
        spec=client.V1PodSpec(
            service_account_name=ACCESS_ALL_SA,
            volumes=[
                client.V1Volume(
                    name="model-data",
                    empty_dir=client.V1EmptyDirVolumeSource(),
                )
            ],
            init_containers=[init_aws(config.aws, model_group), init_parse_manifest()],
            containers=[client.V1Container(**container_args)],
            tolerations=[
                client.V1Toleration(
                    key="app",
                    value="model-group",
                    effect="NoSchedule",
                ),
                client.V1Toleration(
                    key="model",
                    value=model_group.name,
                    effect="NoSchedule",
                ),
            ],
            affinity=client.V1Affinity(
                node_affinity=client.V1NodeAffinity(
                    required_during_scheduling_ignored_during_execution=client.V1NodeSelector(
                        node_selector_terms=[
                            client.V1NodeSelectorTerm(
                                match_expressions=[
                                    client.V1NodeSelectorRequirement(
                                        key="app", operator="In", values=["model-group"]
                                    ),
                                    client.V1NodeSelectorRequirement(
                                        key="model",
                                        operator="In",
                                        values=[model_group.name],
                                    ),
                                ]
                            )
                        ]
                    )
                ),
                pod_anti_affinity=client.V1PodAntiAffinity(
                    required_during_scheduling_ignored_during_execution=[
                        client.V1PodAffinityTerm(
                            label_selector=client.V1LabelSelector(
                                match_labels={
                                    "app": "model-group",
                                    "model": model_group.name,
                                }
                            ),
                            topology_key="kubernetes.io/hostname",
                        )
                    ]
                ),
            ),
        ),
    )


def create_deployment(
    namespace: str, model_group: CloudModelGroup, pod: client.V1Pod
) -> client.V1Deployment:
    """
    Creates a Kubernetes Deployment for a machine learning model group.

    Args:
        namespace (str): The namespace to create the Deployment in.
        model_group (CloudModelGroup): The model group to run in the Deployment.
        pod (client.V1Pod): The Pod to use as a template for the Deployment.

    Returns:
        client.V1Deployment: The created Deployment.
    """
    return client.V1Deployment(
        api_version="apps/v1",
        kind="Deployment",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
        ),
        spec=client.V1DeploymentSpec(
            replicas=model_group.minInstances,
            selector=client.V1LabelSelector(
                match_labels={
                    "app": "model-group",
                    "model": model_group.name,
                }
            ),
            template=pod,
        ),
    )


def create_service_monitor(namespace: str, model_group: CloudModelGroup) -> None:
    monitor = CustomResource(
        api_version="monitoring.coreos.com/v1",
        kind="ServiceMonitor",
        plural="servicemonitors",
        metadata=client.V1ObjectMeta(
            name=f"{model_group.name}-monitor", namespace=namespace
        ),
        spec={
            "selector": {
                "matchLabels": {"app": "model-group", "model": model_group.name}
            },
            "namespaceSelector": {
                "matchNames": [namespace],
            },
            "endpoints": [
                {
                    "port": "http-envoy-prom",
                    "path": "/stats/prometheus",
                    "interval": "15s",
                },
            ],
        },
    )
    apply_resource(monitor)


def create_service(
    namespace: str, model_group: CloudModelGroup, port: int, sidecar_port: int = 15090
) -> client.V1Service:
    """
    Creates a Kubernetes Service for a machine learning model group.

    Args:
        namespace (str): The namespace to create the Service in.
        model_group (CloudModelGroup): The model group to expose with the Service.
        port (int): The port to expose on the Service.
        sidecar_port (int): The port on which the istio sidecar is exposing metrics.


    Returns:
        client.V1Service: The created Service.
    """
    return client.V1Service(
        api_version="v1",
        kind="Service",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
            labels={
                "app": "model-group",
                "model": model_group.name,
            },
        ),
        spec=client.V1ServiceSpec(
            selector={
                "app": "model-group",
                "model": model_group.name,
            },
            ports=[
                client.V1ServicePort(
                    name="http-app",
                    port=80,
                    target_port=port,
                ),
                client.V1ServicePort(
                    name="http-envoy-prom",
                    port=sidecar_port,
                    target_port=sidecar_port,
                ),
            ],
        ),
    )


def filter_services(namespace: str) -> List[Any]:
    """
    Filters the Kubernetes Services in a namespace that belong to a model group.

    Args:
        namespace (str): The namespace to filter the Services in.

    Returns:
        List[Any]: The filtered Services.
    """

    v1 = client.CoreV1Api()

    services = v1.list_namespaced_service(namespace)
    filtered_services = [
        service
        for service in services.items
        if service.spec.selector
        and service.spec.selector.get("app") == "model-group"
        and service.spec.selector.get("model")
    ]

    return filtered_services


def create_hpa(
    namespace: str, model_group: CloudModelGroup, deployment: client.V1Deployment
) -> client.V2HorizontalPodAutoscaler:
    """
    Creates a Kubernetes Horizontal Pod Autoscaler (HPA) for a machine learning model group.

    This function creates a Kubernetes HPA with the specified namespace, model group, and deployment.
    The HPA automatically scales the number of Pods in the deployment based on the CPU utilization.

    The HPA is configured to maintain a specified number of replicas of the Pod.
    The HPA increases the number of replicas when the average CPU utilization exceeds 95%.

    Args:
        namespace (str): The namespace to create the HPA in.
        model_group (CloudModelGroup): The model group to scale with the HPA.
        deployment (client.V1Deployment): The deployment to scale.

    Returns:
        client.V2HorizontalPodAutoscaler: The created HPA.
    """
    return client.V2HorizontalPodAutoscaler(
        api_version="autoscaling/v2",
        kind="HorizontalPodAutoscaler",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
        ),
        spec=client.V2HorizontalPodAutoscalerSpec(
            scale_target_ref=client.V2CrossVersionObjectReference(
                api_version="apps/v1",
                kind="Deployment",
                name=deployment.metadata.name,
            ),
            min_replicas=model_group.minInstances,
            max_replicas=model_group.maxInstances,
            metrics=[
                client.V2MetricSpec(
                    type="Resource",
                    resource=client.V2ResourceMetricSource(
                        name="cpu",
                        target=client.V2MetricTarget(
                            type="Utilization",
                            average_utilization=50,
                        ),
                    ),
                )
            ],
        ),
    )


def create_scaled_object(
    namespace: str, model_group: CloudModelGroup, deployment: client.V1Deployment
) -> Optional[CustomResource]:
    """
    Creates a KEDA ScaledObject for a given model group.

    This function creates a ScaledObject custom resource for Kubernetes Event-driven Autoscaling (KEDA).
    The ScaledObject is used to scale a Kubernetes Deployment based on the triggers defined in the model group.

    Args:
        namespace (str): The namespace in which to create the ScaledObject.
        model_group (CloudModelGroup): The model group for which to create the ScaledObject.
            This object should have `autoScaleTriggers`, `minInstances`, and `maxInstances` attributes.
        deployment (client.V1Deployment): The Kubernetes Deployment that the ScaledObject should scale.

    Returns:
        None
    """
    if not model_group.autoScaleTriggers:
        return None

    return CustomResource(
        api_version="keda.sh/v1alpha1",
        kind="ScaledObject",
        plural="scaledobjects",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}", namespace=namespace
        ),
        spec={
            "scaleTargetRef": {
                "kind": "Deployment",
                "name": deployment.metadata.name,
            },
            "minReplicaCount": model_group.minInstances,
            "maxReplicaCount": model_group.maxInstances,
            "pollingInterval": 15,
            "triggers": list(
                map(
                    lambda trigger: {
                        "type": trigger.type,
                        "metadata": trigger.metadata,
                    },
                    model_group.autoScaleTriggers,
                )
            ),
        },
    )


def create_model_group_service(
    namespace: str,
    config: Config,
    model_group: CloudModelGroup,
) -> None:
    """
    Creates a Kubernetes service for a machine learning model group.

    Args:
        namespace (str): The namespace to create the service in.
        config (Config): The configuration for the service.
        model_group (CloudModelGroup): The model group to create the service for.

    Raises:
        ValueError: If the AWS configuration is not provided.

    Returns:
        None
    """
    if config.aws is None:
        raise ValueError("Only AWS is supported at this time")

    # Download the model to S3 first
    download_model(model_group.name)

    port = 8000

    pod = create_pod(
        namespace,
        config,
        model_group,
        (LLAMA_CPP_PYTHON_CUDA if model_group.awsGpu else LLAMA_CPP_PYTHON_IMAGE),
        port,
    )

    deployment = create_deployment(namespace, model_group, pod)
    apply_resource(deployment)

    svc = create_service(namespace, model_group, port)
    apply_resource(svc)

    if config.aws.prometheus and config.aws.prometheus.enabled:
        create_service_monitor(namespace, model_group)

    scaled_object = create_scaled_object(namespace, model_group, deployment)
    if scaled_object:
        apply_resource(scaled_object)



import os
import platform
import shutil
import tarfile
import tempfile
import zipfile

import requests

from paka.logger import logger


def get_latest_pack_version() -> str:
    url = "https://api.github.com/repos/buildpacks/pack/releases/latest"
    response = requests.get(url)
    response.raise_for_status()
    data = response.json()
    return data["tag_name"][1:]  # Remove the 'v' prefix


def install_pack() -> None:
    if shutil.which("pack"):
        return

    system = platform.system().lower()
    arch = platform.machine().lower()

    if system == "darwin":
        system = "macos"

    pack_version = get_latest_pack_version()

    if system == "windows":
        pack_file = f"pack-v{pack_version}-windows.zip"

        program_files_dir = os.environ.get("PROGRAMFILES", "C:\\Program Files")
        bin_dir = os.path.join(
            program_files_dir, "pack"
        )  # Replace 'YourAppName' with your application's name

        # Check if the script is running with administrator privileges
        if not os.path.exists(bin_dir):
            try:
                os.makedirs(bin_dir)
            except PermissionError:
                logger.error("Administrator privileges required.")
                raise

    elif arch in ["amd64", "x86_64"]:
        pack_file = f"pack-v{pack_version}-{system}.tgz"
    elif arch == "arm64":
        pack_file = f"pack-v{pack_version}-{system}-{arch}.tgz"
    else:
        raise Exception(f"Unsupported architecture: {arch}")

    url = f"https://github.com/buildpacks/pack/releases/download/v{pack_version}/{pack_file}"

    logger.info(f"Downloading {pack_file}...")
    try:
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                for chunk in r.iter_content(chunk_size=8192):
                    tf.write(chunk)
            local_file = tf.name

        if system == "windows":
            with zipfile.ZipFile(local_file, "r") as zip_ref:
                zip_ref.extractall(bin_dir)
        else:
            with tarfile.open(local_file, "r:gz") as tar:
                tar.extractall("/usr/local/bin")
            os.chmod("/usr/local/bin/pack", 0o755)
    finally:
        if local_file is not None:
            os.unlink(local_file)

    logger.info("Pack installed successfully.")



import pulumi_aws as aws
import pulumi_eks as eks

from paka.config import CloudConfig


def odic_role_for_sa(
    config: CloudConfig,
    cluster: eks.Cluster,
    role_name: str,
    ns_service_account: str,
) -> aws.iam.Role:
    """
    Creates an IAM role for a service account in an EKS cluster using OpenID Connect (OIDC) authentication.

    Args:
        config (CloudConfig): The cloud configuration.
        cluster (eks.Cluster): The EKS cluster.
        role_name (str): The name of the role.
        ns_service_account (str): The name of the service account. e.g. "default:sa", "kube-system:auto-scaler"

    Returns:
        aws.iam.Role: The IAM role for the service account.
    """
    project = config.cluster.name

    role = aws.iam.Role(
        f"{project}-{role_name}-role",
        assume_role_policy=cluster.core.oidc_provider.url.apply(
            lambda url: aws.iam.get_policy_document(
                statements=[
                    aws.iam.GetPolicyDocumentStatementArgs(
                        effect="Allow",
                        principals=[
                            aws.iam.GetPolicyDocumentStatementPrincipalArgs(
                                type="Federated",
                                identifiers=[cluster.core.oidc_provider.arn],
                            )
                        ],
                        actions=["sts:AssumeRoleWithWebIdentity"],
                        conditions=[
                            aws.iam.GetPolicyDocumentStatementConditionArgs(
                                test="StringEquals",
                                variable=f"{url}:sub",
                                values=[f"system:serviceaccount:{ns_service_account}"],
                            )
                        ],
                    )
                ],
            ).json
        ),
    )

    return role



import logging

# Create a logger
logger = logging.getLogger(__name__)


# "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
def setup_logger(verbose: bool = False, format: str = "%(message)s") -> None:
    # Set the logging level based on the verbose flag
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)

    for handler in logger.handlers:
        logger.removeHandler(handler)

    # Create a console handler
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG if verbose else logging.INFO)

    # Create a formatter
    formatter = logging.Formatter(format)

    # Add the formatter to the console handler
    ch.setFormatter(formatter)

    # Add the console handler to the logger
    logger.addHandler(ch)


setup_logger()



import os
from typing import Optional

import click
import typer
from kubernetes import client

from paka.cli.utils import resolve_image
from paka.k8s import try_load_kubeconfig
from paka.kube_resources.job.worker import create_workers, delete_workers
from paka.logger import logger
from paka.utils import kubify_name, read_current_cluster_data

try_load_kubeconfig()

job_app = typer.Typer()


def prefixed_job_name(job_name: str) -> str:
    # Prefix the job name with "job-" to reduce the chances of name collision
    # between jobs and functions.
    if not job_name.startswith("job-"):
        return f"job-{job_name}"
    return job_name


@job_app.command()
def deploy(
    name: str = typer.Option(
        "",
        "--name",
        help="The name of the job. This name will be used to track the job's "
        "progress and manage its resources. If not provided, the system will use "
        "the base name of the 'source_dir' or the 'image' name as the job name.",
    ),
    entrypoint: str = typer.Option(
        ...,
        "--entrypoint",
        help="The entrypoint of the application. This refers to the command "
        "defined in the Procfile that will be executed to start the application "
        "when the job is run.",
    ),
    source_dir: Optional[str] = typer.Option(
        None,
        "--source",
        help="The directory containing the source code of the application. If "
        "specified, a new Docker image will be built using the source code from "
        "this directory. A Dockerfile is not required because the build process "
        "uses Cloud Native's Buildpacks, which automatically detect and install "
        "dependencies.",
    ),
    image: Optional[str] = typer.Option(
        None,
        "--image",
        help="The name of the Docker image to deploy. If both an image and a "
        "source directory are provided, this image will be used and the source "
        "directory will be ignored.",
    ),
    max_workers: int = typer.Option(
        5,
        "--max-workers",
        help="The maximum number of workers that can be allocated for this job. "
        "This sets an upper limit on the number of workers that the autoscaling "
        "system can create to handle tasks for this job.",
    ),
    tasks_per_worker: int = typer.Option(
        5,
        "--tasks-per-worker",
        help="The number of tasks that need to be allocated to each worker before "
        "the autoscaler creates a new worker.",
    ),
    wait_existing_tasks: bool = typer.Option(
        True,
        "--wait-existing-tasks",
        help="Determines whether the system should wait for existing tasks to "
        "complete before deploying the new job. If set to true, the deployment "
        "will wait until all current tasks have finished.",
    ),
) -> None:
    """
    Deploy a job.

    A job leverages a Redis queue to distribute work to a pool of autoscaling workers.
    The job can be specified by providing a Docker image or a source directory containing the application code.

    If a Docker image is provided, the cluster will run the job using this image.
    If a source directory is provided, the cluster will build a Docker image from the source code and then run the job.

    If both an image and a source directory are provided, the Docker image is used and the source directory is ignored.
    """
    resolved_image = resolve_image(image, source_dir)

    if image:
        job_name = image
    elif source_dir:
        job_name = os.path.basename(source_dir)

    create_workers(
        namespace=read_current_cluster_data("namespace"),
        job_name=kubify_name(prefixed_job_name(name or job_name)),
        image=resolved_image,
        entrypoint=entrypoint,
        tasks_per_worker=tasks_per_worker,
        max_replicas=max_workers,
        drain_existing_job=wait_existing_tasks,
    )


@job_app.command()
def delete(
    name: str = typer.Argument(
        ...,
        help="The name of the job to delete.",
    ),
    wait_existing_tasks: bool = typer.Option(
        True,
        "--wait-existing-tasks",
        help="Determines whether the system should wait for existing tasks to "
        "complete before deploying the new job. If set to true, the deployment "
        "will wait until all current tasks have finished.",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the deletion.",
    ),
) -> None:
    """
    Deletes a job.

    Args:
        name (str): The unique identifier of the job to be deleted.
        wait_existing_tasks (bool): Determines whether the system should wait for existing tasks to complete before deleting the job.
        yes (bool): If True, bypasses the confirmation prompt and directly proceeds with the deletion.

    Returns:
        None
    """
    if yes or click.confirm(
        f"Are you sure you want to delete the job {name}?", default=False
    ):
        logger.info(f"Deleting job {name}")
        delete_workers(
            read_current_cluster_data("namespace"),
            prefixed_job_name(name),
            wait_existing_tasks,
        )
        logger.info(f"Successfully deleted job {name}")


@job_app.command()
def list() -> None:
    """
    Lists all jobs.
    """
    api_instance = client.AppsV1Api()

    label_selector = "role=worker"

    # List the deployments in the specified namespace that match the field selector
    deployments = api_instance.list_namespaced_deployment(
        namespace=read_current_cluster_data("namespace"), label_selector=label_selector
    )

    for deployment in deployments.items:
        logger.info(deployment.metadata.name)

    if not deployments.items:
        logger.info("No jobs found.")



import pulumi
import pulumi_kubernetes as k8s
from pulumi_kubernetes.helm.v3 import Chart, ChartOpts, FetchOpts

from paka.config import CloudConfig
from paka.utils import call_once


@call_once
def create_qdrant(
    config: CloudConfig,
    k8s_provider: k8s.Provider,
) -> None:
    """
    Installs the qdrant helm chart.
    """
    if not config.vectorStore:
        return

    ns = k8s.core.v1.Namespace(
        "qdrant",
        metadata={"name": "qdrant", "labels": {"istio-injection": "enabled"}},
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )

    resource_request = (
        {
            "resources": {
                "requests": {
                    "cpu": config.vectorStore.resourceRequest.cpu,
                    "memory": config.vectorStore.resourceRequest.memory,
                },
            }
        }
        if config.vectorStore.resourceRequest
        else {}
    )

    Chart(
        "qdrant",
        ChartOpts(
            chart="qdrant",
            version="0.7.5",
            namespace="qdrant",
            fetch_opts=FetchOpts(repo="https://qdrant.github.io/qdrant-helm"),
            values={
                "metrics": {
                    "serviceMonitor": {
                        "enabled": (
                            True
                            if config.prometheus and config.prometheus.enabled
                            else False
                        ),
                    },
                },
                "replicaCount": config.vectorStore.replicas,
                "persistence": {
                    "size": config.vectorStore.storage_size,
                },
                "livenessProbe": {
                    "enabled": True,
                },
                "tolerations": [
                    {
                        "key": "app",
                        "operator": "Equal",
                        "value": "qdrant",
                        "effect": "NoSchedule",
                    }
                ],
                "affinity": {
                    "nodeAffinity": {
                        "requiredDuringSchedulingIgnoredDuringExecution": {
                            "nodeSelectorTerms": [
                                {
                                    "matchExpressions": [
                                        {
                                            "key": "app",
                                            "operator": "In",
                                            "values": ["qdrant"],
                                        }
                                    ]
                                }
                            ]
                        }
                    },
                    "podAntiAffinity": {
                        "requiredDuringSchedulingIgnoredDuringExecution": [
                            {
                                "labelSelector": {
                                    "matchExpressions": [
                                        {
                                            "key": "app",
                                            "operator": "In",
                                            "values": ["qdrant"],
                                        }
                                    ]
                                },
                                "topologyKey": "kubernetes.io/hostname",
                            }
                        ]
                    },
                },
                "topologySpreadConstraints": [
                    {
                        "maxSkew": 1,
                        "topologyKey": "topology.kubernetes.io/zone",
                        "whenUnsatisfiable": "ScheduleAnyway",
                        "labelSelector": {"matchLabels": {"app": "qdrant"}},
                    }
                ],
                **resource_request,
            },
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[ns]),
    )



import pulumi
import pulumi_aws as aws
import pulumi_eks as eks
import pulumi_kubernetes as k8s
from pulumi_kubernetes.helm.v3 import Chart, ChartOpts, FetchOpts

from paka.cluster.aws.utils import odic_role_for_sa
from paka.config import CloudConfig
from paka.utils import call_once


@call_once
def create_ebs_csi_driver(
    config: CloudConfig, cluster: eks.Cluster, k8s_provider: k8s.Provider
) -> None:
    project = config.cluster.name

    csi_driver_policy_doc = aws.iam.get_policy_document(
        statements=[
            aws.iam.GetPolicyDocumentStatementArgs(
                actions=[
                    "ec2:CreateSnapshot",
                    "ec2:AttachVolume",
                    "ec2:DetachVolume",
                    "ec2:ModifyVolume",
                    "ec2:DescribeAvailabilityZones",
                    "ec2:DescribeInstances",
                    "ec2:DescribeSnapshots",
                    "ec2:DescribeTags",
                    "ec2:DescribeVolumes",
                    "ec2:DescribeVolumesModifications",
                    "ec2:CreateTags",
                    "ec2:CreateVolume",
                    "ec2:DeleteVolume",
                ],
                resources=["*"],
                effect="Allow",
            )
        ]
    )

    csi_driver_policy = aws.iam.Policy(
        f"{project}-csi-driver-policy", policy=csi_driver_policy_doc.json
    )

    csi_driver_role = odic_role_for_sa(
        config, cluster, "csi-driver", "kube-system:ebs-csi-controller-sa"
    )

    aws.iam.RolePolicyAttachment(
        f"{project}-csi-driver-role-policy-attachment",
        policy_arn=csi_driver_policy.arn,
        role=csi_driver_role.name,
    )

    Chart(
        "aws-ebs-csi-driver",
        ChartOpts(
            chart="aws-ebs-csi-driver",
            version="2.26.0",
            namespace="kube-system",
            fetch_opts=FetchOpts(
                repo="https://kubernetes-sigs.github.io/aws-ebs-csi-driver"
            ),
            values={
                "controller": {
                    "serviceAccount": {
                        "create": "true",
                        "name": "ebs-csi-controller-sa",
                        "annotations": {
                            "eks.amazonaws.com/role-arn": csi_driver_role.arn
                        },
                        "automountServiceAccountToken": "true",
                    },
                }
            },
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )



import pulumi
import pulumi_kubernetes as k8s

from paka.constants import ACCESS_ALL_SA
from paka.utils import call_once, read_current_cluster_data


@call_once
def create_fluentbit(fluent_bit_config: str, k8s_provider: k8s.Provider) -> None:
    """
    Creates a fluentbit daemonset with the given configuration.
    """

    parsers_config = """
[PARSER]
    Name        docker
    Format      json
    Time_Key    time
    Time_Format %Y-%m-%dT%H:%M:%S.%fZ
"""
    parsers_config_map = k8s.core.v1.ConfigMap(
        "fluent-bit-parsers",
        data={"parsers.conf": parsers_config},
        metadata={
            "namespace": read_current_cluster_data("namespace"),
            "name": "fluent-bit-parsers",
        },
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )

    fluent_bit_config_map = k8s.core.v1.ConfigMap(
        "fluent-bit-config-map",
        data={"fluent-bit.conf": fluent_bit_config},
        metadata={
            "namespace": read_current_cluster_data("namespace"),
            "name": "fluent-bit-config",
        },
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )

    k8s.apps.v1.DaemonSet(
        "fluent-bit-daemonset",
        spec=k8s.apps.v1.DaemonSetSpecArgs(
            selector=k8s.meta.v1.LabelSelectorArgs(
                match_labels={"k8s-app": "fluent-bit-logging"},
            ),
            template=k8s.core.v1.PodTemplateSpecArgs(
                metadata=k8s.meta.v1.ObjectMetaArgs(
                    labels={"k8s-app": "fluent-bit-logging"},
                    annotations={"sidecar.istio.io/inject": "false"},
                ),
                spec=k8s.core.v1.PodSpecArgs(
                    service_account_name=ACCESS_ALL_SA,
                    tolerations=[k8s.core.v1.TolerationArgs(operator="Exists")],
                    containers=[
                        k8s.core.v1.ContainerArgs(
                            name="fluent-bit",
                            image="fluent/fluent-bit:latest",
                            volume_mounts=[
                                k8s.core.v1.VolumeMountArgs(
                                    name="config",
                                    mount_path="/fluent-bit/etc/fluent-bit.conf",
                                    sub_path="fluent-bit.conf",
                                ),
                                k8s.core.v1.VolumeMountArgs(
                                    name="varlog",
                                    mount_path="/var/log",
                                ),
                                k8s.core.v1.VolumeMountArgs(
                                    name="varlibdockercontainers",
                                    mount_path="/var/lib/docker/containers",
                                    read_only=True,
                                ),
                                k8s.core.v1.VolumeMountArgs(
                                    name="parsers-config",
                                    mount_path="/fluent-bit/etc/parsers.conf",
                                    sub_path="parsers.conf",
                                ),
                            ],
                        )
                    ],
                    volumes=[
                        k8s.core.v1.VolumeArgs(
                            name="config",
                            config_map=k8s.core.v1.ConfigMapVolumeSourceArgs(
                                name=fluent_bit_config_map.metadata["name"],
                            ),
                        ),
                        k8s.core.v1.VolumeArgs(
                            name="varlog",
                            host_path=k8s.core.v1.HostPathVolumeSourceArgs(
                                path="/var/log",
                            ),
                        ),
                        k8s.core.v1.VolumeArgs(
                            name="varlibdockercontainers",
                            host_path=k8s.core.v1.HostPathVolumeSourceArgs(
                                path="/var/lib/docker/containers",
                            ),
                        ),
                        k8s.core.v1.VolumeArgs(
                            name="parsers-config",
                            config_map=k8s.core.v1.ConfigMapVolumeSourceArgs(
                                name=parsers_config_map.metadata["name"],
                            ),
                        ),
                    ],
                ),
            ),
        ),
        metadata={"namespace": read_current_cluster_data("namespace")},
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )



from typing import Any, Callable, Dict, Optional, Tuple

import pulumi
import pulumi_kubernetes as k8s
from pulumi_kubernetes.helm.v3 import Chart, ChartOpts, FetchOpts

from paka.config import CloudConfig


def memoize(func: Callable[..., Any]) -> Callable[..., Any]:
    cache: Dict[Callable[..., Any], Any] = dict()

    def memoized_func(*args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> Any:
        if func not in cache:
            cache[func] = func(*args, **kwargs)
        return cache[func]

    return memoized_func


@memoize
def create_prometheus(
    config: CloudConfig, k8s_provider: k8s.Provider
) -> Optional[Chart]:
    """
    Installs a Prometheus chart.
    """
    if not config.prometheus or not config.prometheus.enabled:
        return None

    ns = k8s.core.v1.Namespace(
        "prometheus",
        metadata={"name": "prometheus"},
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )

    return Chart(
        "kube-prometheus-stack",
        ChartOpts(
            chart="kube-prometheus-stack",
            version="56.3.0",
            namespace="prometheus",
            fetch_opts=FetchOpts(
                repo="https://prometheus-community.github.io/helm-charts"
            ),
            values={
                "nodeExporter": {
                    "enabled": config.prometheus.node_exporter,
                },
                "alertmanager": {
                    "enabled": config.prometheus.alertmanager,
                },
                "grafana": {
                    "enabled": config.prometheus.grafana,
                },
                "kubeApiServer": {
                    "enabled": config.prometheus.kube_api_server,
                },
                "kubelet": {
                    "enabled": config.prometheus.kubelet,
                },
                "kubeControllerManager": {
                    "enabled": config.prometheus.kube_controller_manager,
                },
                "coreDns": {
                    "enabled": config.prometheus.core_dns,
                },
                "kubeEtcd": {
                    "enabled": config.prometheus.kube_etcd,
                },
                "kubeScheduler": {
                    "enabled": config.prometheus.kube_scheduler,
                },
                "kubeProxy": {
                    "enabled": config.prometheus.kube_proxy,
                },
                "kubeStateMetrics": {
                    "enabled": config.prometheus.kube_state_metrics,
                },
                "thanosRuler": {
                    "enabled": config.prometheus.thanos_ruler,
                },
                # Disable the Prometheus Operator's admission webhooks, since they don't work with Pulumi.
                # This means ill-formatted Prometheus rules may make their way into Prometheus. :(
                "prometheusOperator": {
                    "admissionWebhooks": {"enabled": False},
                    "tls": {"enabled": False},
                },
                "kube-state-metrics": {
                    "metricLabelsAllowlist": [
                        "pods=[*]",
                        "deployments=[app.kubernetes.io/name,app.kubernetes.io/component,app.kubernetes.io/instance]",
                    ]
                },
                "prometheus": {
                    "prometheusSpec": {
                        "serviceMonitorSelectorNilUsesHelmValues": False,
                        "podMonitorSelectorNilUsesHelmValues": False,
                        "storageSpec": {
                            "volumeClaimTemplate": {
                                "spec": {
                                    "accessModes": ["ReadWriteOnce"],
                                    "resources": {
                                        "requests": {
                                            "storage": config.prometheus.storage_size,
                                        }
                                    },
                                }
                            }
                        },
                    }
                },
            },
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[ns]),
    )



import pulumi_aws as aws

from paka.config import CloudConfig
from paka.utils import call_once, save_cluster_data


@call_once
def create_object_store(config: CloudConfig) -> None:
    """
    Creates an object store in AWS S3 based on the provided configuration.

    Args:
        config (CloudConfig): The configuration object containing cluster information.

    Returns:
        None
    """
    # `bucket` is the name of the bucket. It will avoid pulumi appending a random string to the name
    # `force_destroy`` is needed to delete the bucket when it's not empty
    bucket = aws.s3.Bucket(config.cluster.name, force_destroy=True)
    bucket.id.apply(lambda id: save_cluster_data(config.cluster.name, "bucket", id))



from functools import partial
from typing import Any, Callable, Dict

import pulumi
import pulumi_kubernetes as k8s
from pulumi import ResourceOptions
from pulumi_kubernetes.yaml import ConfigFile

from paka.cluster.prometheus import create_prometheus
from paka.config import CloudConfig
from paka.utils import call_once

VERSION = "v1.12.3"
ISTIO_VERSION = "v1.12.1"


def limit_hpa_min_replicas(args: Any, opts: pulumi.ResourceOptions) -> None:
    if (
        args["kind"] == "HorizontalPodAutoscaler"
        and args["metadata"]["name"] == "istiod"
    ):
        args["spec"]["minReplicas"] = 1


def limit_resources(args: Any, opts: pulumi.ResourceOptions) -> None:
    if args["kind"] == "Deployment" and args["metadata"]["name"] == "istiod":
        for container in args["spec"]["template"]["spec"]["containers"]:
            if container["name"] == "discovery":
                container["resources"] = {
                    "requests": {"cpu": "300m", "memory": "1Gi"},
                }


def limit_deployment_replicas(args: Any, opts: pulumi.ResourceOptions) -> None:
    if (
        args["kind"] == "Deployment"
        and args["metadata"]["name"] == "istio-ingressgateway"
    ):
        args["spec"]["replicas"] = 2
        for container in args["spec"]["template"]["spec"]["containers"]:
            if container["name"] == "istio-proxy":
                container["resources"] = {
                    "requests": {"cpu": "500m", "memory": "500Mi"},
                }


def crd_resources(labels: Dict[str, Any]) -> bool:
    return labels.get("knative.dev/crd-install") == "true"


def non_crd_resources(labels: Dict[str, Any]) -> bool:
    return labels.get("knative.dev/crd-install") != "true"


def crd_install_filter(
    inputs: Any, opts: ResourceOptions, filter: Callable[[Any], bool]
) -> None:
    if "metadata" in inputs:
        if not filter(inputs["metadata"].get("labels", {})):
            inputs["kind"] = "List"
            inputs["items"] = []


def exclude_knative_eventing_namespace(inputs: Any, opts: ResourceOptions) -> None:
    if "metadata" in inputs and inputs["metadata"]["namespace"] == "knative-eventing":
        inputs["kind"] = "List"
        inputs["items"] = []


only_crd_transform = partial(crd_install_filter, filter=crd_resources)
non_crd_transform = partial(crd_install_filter, filter=non_crd_resources)


# TODO: Decouple knative and istio
@call_once
def create_knative_and_istio(config: CloudConfig, k8s_provider: k8s.Provider) -> None:
    ns = k8s.core.v1.Namespace(
        "knative-serving",
        metadata={"name": "knative-serving"},
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )

    yaml_files = [
        # TODO: sigstore verification
        # Creates resources under the knative-serving namespace
        f"https://github.com/knative/serving/releases/download/knative-{VERSION}/serving-core.yaml",
    ]
    for i, yaml_file in enumerate(yaml_files):
        ConfigFile(
            yaml_file.split("/")[-1],
            file=yaml_file,
            opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[ns]),
        )

    yaml_file = f"https://github.com/knative/net-istio/releases/download/knative-{ISTIO_VERSION}/istio.yaml"
    istio_crd_install = ConfigFile(
        "istio-crd-install",
        file=yaml_file,
        transformations=[only_crd_transform],
        opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[ns]),
    )

    istio_full_install = ConfigFile(
        "istio-non-crd-install",
        file=yaml_file,
        transformations=[
            non_crd_transform,
            limit_resources,
            limit_hpa_min_replicas,
            limit_deployment_replicas,
        ],
        opts=pulumi.ResourceOptions(
            provider=k8s_provider, depends_on=[istio_crd_install, ns]
        ),
    )

    yaml_file = f"https://github.com/knative/net-istio/releases/download/knative-{ISTIO_VERSION}/net-istio.yaml"
    net_istio = ConfigFile(
        "net-istio",
        file=yaml_file,
        opts=pulumi.ResourceOptions(
            provider=k8s_provider, depends_on=[istio_full_install, ns]
        ),
    )

    yaml_file = f"https://github.com/knative/serving/releases/download/knative-{VERSION}/serving-default-domain.yaml"
    ConfigFile(
        "kn-default-domain",
        file=yaml_file,
        opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[net_istio, ns]),
    )

    # Now, install ServiceMonitor and PodMonitor CRDs
    prometheus = create_prometheus(config, k8s_provider)

    if not prometheus:
        return

    yaml_file = "https://raw.githubusercontent.com/knative-extensions/monitoring/main/servicemonitor.yaml"

    ConfigFile(
        "kn-prom-monitor",
        file=yaml_file,
        transformations=[exclude_knative_eventing_namespace],
        opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[prometheus, ns]),
    )



from typing import List

from kubernetes import client

from paka.k8s import CustomResource, apply_resource
from paka.utils import kubify_name


def create_model_group_ingress(namespace: str) -> None:
    istio_gateway_crd = CustomResource(
        api_version="networking.istio.io/v1beta1",
        kind="Gateway",
        plural="gateways",
        metadata=client.V1ObjectMeta(
            name="model-group-ingress-gateway", namespace=namespace
        ),
        spec={
            "selector": {
                "istio": "ingressgateway"  # Use Istio's default ingress gateway
            },
            "servers": [
                {
                    "port": {"number": 80, "name": "http", "protocol": "HTTP"},
                    "hosts": ["*"],
                }
            ],
        },
    )

    apply_resource(istio_gateway_crd)


def create_model_vservice(
    namespace: str, model_name: str, hosts: List[str] = ["*"]
) -> None:
    istio_virtual_service = CustomResource(
        api_version="networking.istio.io/v1beta1",
        kind="VirtualService",
        plural="virtualservices",
        metadata=client.V1ObjectMeta(name=kubify_name(model_name), namespace=namespace),
        spec={
            "hosts": hosts,
            "gateways": ["model-group-ingress-gateway"],
            "http": [
                {
                    "match": [
                        {
                            "authority": {
                                "prefix": kubify_name(model_name),
                            }
                        }
                    ],
                    "route": [
                        {
                            "destination": {
                                "host": f"{kubify_name(model_name)}.{namespace}.svc.cluster.local",
                                "port": {"number": 80},
                            }
                        }
                    ],
                }
            ],
        },
    )

    apply_resource(istio_virtual_service)



import pulumi
import pulumi_kubernetes as k8s
from pulumi_kubernetes.apiextensions import CustomResource
from pulumi_kubernetes.helm.v3 import Chart, ChartOpts, FetchOpts

from paka.config import CloudConfig
from paka.utils import call_once, read_current_cluster_data


@call_once
def create_redis(config: CloudConfig, k8s_provider: k8s.Provider) -> None:
    """
    Installs redis with a helm chart.
    """
    if not config.job or not config.job.enabled:
        return

    ns = k8s.core.v1.Namespace(
        "redis",
        metadata={"name": "redis"},
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )

    chart = Chart(
        "redis",
        ChartOpts(
            chart="redis",
            version="18.6.1",
            namespace=read_current_cluster_data("namespace"),
            fetch_opts=FetchOpts(repo="https://charts.bitnami.com/bitnami"),
            values={
                "architecture": "standalone",
                "master": {
                    "persistence": {
                        "enabled": True,
                        "size": config.job.broker_storage_size,
                    },
                },
                "metrics": {"enabled": True},  # For enabling metrics
            },
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider, depends_on=[ns]),
    )

    if not config.prometheus or not config.prometheus.enabled:
        return

    CustomResource(
        "redis-metrics-monitor",
        api_version="monitoring.coreos.com/v1",
        kind="ServiceMonitor",
        metadata={
            "name": "redis-metrics-monitor",
            "namespace": "redis",
        },
        spec={
            "selector": {
                "matchLabels": {
                    "app.kubernetes.io/instance": "redis",
                    "app.kubernetes.io/name": "redis",
                    "app.kubernetes.io/component": "metrics",
                }
            },
            "namespaceSelector": {
                "matchNames": ["redis"],
            },
            "endpoints": [
                {
                    "port": "http-metrics",
                    "interval": "15s",
                },
            ],
        },
        opts=pulumi.ResourceOptions(
            provider=k8s_provider,
            depends_on=[chart],
        ),
    )



import shlex
from typing import Optional

import typer
from kubernetes import client

from paka.cli.utils import resolve_image
from paka.k8s import tail_logs, try_load_kubeconfig
from paka.logger import logger
from paka.utils import kubify_name, random_str, read_current_cluster_data

CLEANUP_TIMEOUT = 600  # 10 minutes

run_app = typer.Typer()

try_load_kubeconfig()


@run_app.callback(invoke_without_command=True)
def one_off_script(
    entrypoint: str = typer.Option(
        ...,
        "--entrypoint",
        help="The entrypoint of the application. This refers to the command "
        "defined in the Procfile that will be executed.",
    ),
    source_dir: Optional[str] = typer.Option(
        None,
        "--source",
        help="The directory containing the source code of the application. If "
        "specified, a new Docker image will be built using the source code from "
        "this directory. A Dockerfile is not required because the build process "
        "uses Cloud Native's Buildpacks, which automatically detect and install "
        "dependencies.",
    ),
    image: Optional[str] = typer.Option(
        None,
        "--image",
        help="The name of the Docker image to deploy. If both an image and a "
        "source directory are provided, this image will be used and the source "
        "directory will be ignored.",
    ),
) -> None:
    """
    Runs a one-off script.

    This command creates a new Kubernetes job that runs the specified entrypoint command
    in a container with the specified Docker image. If a source directory is provided, a new
    Docker image is built using the source code from that directory.
    """
    resolved_image = resolve_image(image, source_dir)

    # Generate a job name which is the hash of the command
    job_name = f"run-{kubify_name(random_str(10))}"

    job = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(name=job_name, labels={"job-name": job_name}),
        spec=client.V1JobSpec(
            template=client.V1PodTemplateSpec(
                spec=client.V1PodSpec(
                    containers=[
                        client.V1Container(
                            name="one-off-script",
                            image=resolved_image,
                            image_pull_policy="Always",
                            command=shlex.split(entrypoint),
                        )
                    ],
                    restart_policy="Never",
                )
            ),
            backoff_limit=0,
            ttl_seconds_after_finished=CLEANUP_TIMEOUT,
        ),
    )

    namespace = read_current_cluster_data("namespace")

    logger.info(f"Submitting the task...")
    batch_api = client.BatchV1Api()
    batch_api.create_namespaced_job(namespace=namespace, body=job)
    logger.info(f"Successfully submitted the task.")

    logger.info(f"Waiting for the task to complete...")
    api = client.CoreV1Api()
    pods = api.list_namespaced_pod(
        namespace=namespace, label_selector=f"job-name={job_name}"
    )
    for pod in pods.items:
        tail_logs(namespace, pod.metadata.name, "one-off-script")



import json
from typing import Optional

import boto3
from kubernetes import client, config


# Pulumi cannot update the idle timeout of an ELB. This script uses boto3 to
# update the idle timeout of an ELB.
def _update_elb_idle_timeout(
    load_balancer_name: str, idle_timeout_seconds: int
) -> None:
    elb_client = boto3.client("elb")

    attributes = {
        "LoadBalancerAttributes": {
            "ConnectionSettings": {"IdleTimeout": idle_timeout_seconds}
        }
    }

    elb_client.modify_load_balancer_attributes(
        LoadBalancerName=load_balancer_name,
        LoadBalancerAttributes=attributes["LoadBalancerAttributes"],
    )


def update_elb_idle_timeout(kubeconfig_json: str, idle_timeout_seconds: int) -> None:
    elb_name = get_elb_name(kubeconfig_json)

    if elb_name:
        _update_elb_idle_timeout(elb_name, idle_timeout_seconds)


def get_elb_name(kubeconfig_json: str) -> Optional[str]:
    config.load_kube_config_from_dict(json.loads(kubeconfig_json))

    v1 = client.CoreV1Api()
    services = v1.list_service_for_all_namespaces(watch=False)

    for service in services.items:
        if service.spec.type == "LoadBalancer":
            # The name of the ELB is the first part of the hostname of the load balancer
            elb_hostname = service.status.load_balancer.ingress[0].hostname
            elb_name = elb_hostname.split("-")[0]
            return elb_name

    return None



import functools
import os
import re
import subprocess
from typing import Any, Optional

import typer

from paka.cluster.manager.aws import AWSClusterManager
from paka.cluster.manager.base import ClusterManager
from paka.config import parse_yaml
from paka.constants import BP_BUILDER_ENV_VAR
from paka.container.ecr import push_to_ecr
from paka.container.pack import install_pack
from paka.logger import logger
from paka.utils import get_pulumi_root, read_current_cluster_data


def build(
    source_dir: str,
    image_name: str,
) -> str:
    """
    Builds a Docker image from the source code in the specified directory.

    This function uses the Cloud Native Buildpacks (pack) tool to build a Docker image from the source code
    in the specified directory. The resulting image is then pushed to the Amazon Elastic Container Registry (ECR).

    Args:
        source_dir (str): The directory containing the source code of the application.
        image_name (str): The name of the Docker image to build. If not provided, the name of the source directory is used.

    Returns:
        str: The image tag of the built Docker image.
    """
    install_pack()

    source_dir = os.path.abspath(os.path.expanduser(source_dir))

    if not os.path.exists(os.path.join(source_dir, ".cnignore")):
        logger.error(".cnignore file does not exist in the source directory.")
        raise typer.Exit(1)

    if not os.path.exists(os.path.join(source_dir, "Procfile")):
        logger.error("Procfile does not exist in the source directory.")
        raise typer.Exit(1)

    # If image_name is empty, use the directory name of source_dir
    if not image_name:
        image_name = os.path.basename(source_dir)

    logger.info(f"Building image {image_name}...")

    # Parse runtime.txt and set BP_CPYTHON_VERSION
    runtime_path = os.path.join(source_dir, "runtime.txt")
    bp_cpython_version = None
    bp_node_version = None
    if os.path.exists(runtime_path):
        with open(runtime_path, "r") as file:
            content = file.read()
            match = re.search(r"python-(\d+\.\d+(\.\d+|\.\*))", content)
            if match:
                bp_cpython_version = match.group(1)
            else:
                match = re.search(r"nodejs-(\d+\.\d+(\.\d+|\.\*))", content)
                if match:
                    bp_node_version = match.group(1)

    try:
        subprocess.run(["cd", source_dir], check=True)

        builder = os.environ.get(BP_BUILDER_ENV_VAR, "paketobuildpacks/builder:base")

        pack_command = [
            "pack",
            "build",
            image_name,
            "--builder",
            builder,
        ]
        if bp_cpython_version:
            pack_command.extend(["--env", f"BP_CPYTHON_VERSION={bp_cpython_version}"])
        elif bp_node_version:
            pack_command.extend(["--env", f"BP_NODE_VERSION={bp_node_version}"])

        subprocess.run(pack_command, check=True)
        logger.info(f"Successfully built {image_name}")

        return push_to_ecr(
            image_name,
            read_current_cluster_data("registry"),
            read_current_cluster_data("region"),
            image_name,
        )

    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred: {e}")
        raise typer.Exit(1)


def load_cluster_manager(cluster_config: str) -> ClusterManager:
    """
    Loads the cluster manager based on the provided cluster configuration YAML file.

    This function reads the cluster configuration file, parses it as YAML, and then
    creates and returns a cluster manager object based on the cloud provider specified
    in the configuration. Currently, only AWS is supported.

    Args:
        cluster_config (str): The path to the cluster configuration file. If not
        provided, defaults to "./cluster.yaml".

    Returns:
        ClusterManager: An instance of the cluster manager corresponding to the
        cloud provider specified in the configuration.

    Raises:
        FileNotFoundError: If the cluster configuration file does not exist.
        ValueError: If the cloud provider specified in the configuration is not supported.
    """
    if not cluster_config:
        cluster_config = "./cluster.yaml"

    cluster_config = os.path.abspath(os.path.expanduser(cluster_config))

    if not os.path.exists(cluster_config):
        raise FileNotFoundError(f"The cluster config file does not exist")

    with open(cluster_config, "r") as file:
        config_data = parse_yaml(file.read())

        if config_data.aws:
            return AWSClusterManager(config=config_data)
        else:
            raise ValueError("Unsupported cloud provider")


def validate_name(func: Any) -> Any:
    """
    Decorator to validate the name argument of a function.

    This decorator checks if the name argument of the decorated function matches
    the specified regular expression and is not longer than 63 characters. If
    the name is invalid, it logs an error message and exits the program.

    Args:
        func (Any): The function to decorate.

    Returns:
        Any: The decorated function.
    """

    @functools.wraps(func)
    def wrapper(name: str, *args: Any, **kwargs: Any) -> Any:
        if (
            not re.match(
                r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?$",
                name,
            )
            or len(name) > 63
        ):
            logger.info(
                "Invalid name. It must contain no more than 63 characters, contain "
                "only lowercase alphanumeric characters or '-', start with an "
                "alphanumeric character, and end with an alphanumeric character."
            )
            raise typer.Exit(1)
        return func(name, *args, **kwargs)

    return wrapper


def resolve_image(image: Optional[str], source_dir: Optional[str]) -> str:
    """
    Determines the Docker image to be utilized by either using the provided image
    directly or by building an image from the specified source directory.

    This function checks if exactly one of image or source_dir is provided. If
    both or neither are provided, it logs an error message and exits the program.
    If only source_dir is provided, it builds the Docker image from the source
    directory and sets the image name to the base name of the source directory.
    If only image is provided, it uses the provided image name. If the image
    name does not contain a tag, it adds the default tag 'latest'.

    Args:
        image (Optional[str]): The name of the Docker image to use. If not
            provided, a Docker image is built from the source directory.
        source_dir (Optional[str]): The path to the source directory to build the
            Docker image from. If not provided, the provided image name is used.

    Returns:
        str: The resolved Docker image name.

    Raises:
        typer.Exit: If both or neither image and source_dir are provided.
    """
    if bool(source_dir) == bool(image):
        logger.error(
            "Exactly one of --source or --image must be provided. Please see --help for more information."
        )
        raise typer.Exit(1)

    result_image = ""

    if not image and source_dir:
        source_dir = os.path.abspath(os.path.expanduser(source_dir))
        result_image = os.path.basename(source_dir)
        result_image = build(source_dir, result_image)
    elif image:
        result_image = image

    # If a fully qualified image name is provided, use it directly. This gives users
    # the flexibility to use images from other registries. Otherwise, the image is assumed
    # to be located in the default registry and will be retrieved from there.
    if ":" not in result_image:
        registry_uri = read_current_cluster_data("registry")
        result_image = f"{registry_uri}:{result_image}"

    return result_image


def init_pulumi() -> None:
    """
    Initializes Pulumi by setting up necessary environment variables and creating
    the Pulumi state root directory.

    This function sets the Pulumi configuration passphrase and backend URL
    environment variables. If these variables are not already set in the
    environment, it assigns them default values. The default value for the
    passphrase is an empty string, and the default for the backend URL is a
    file URL pointing to the Pulumi root directory.

    The Pulumi root directory is created if it does not already exist.
    """
    os.environ["PULUMI_CONFIG_PASSPHRASE"] = os.environ.get(
        "PULUMI_CONFIG_PASSPHRASE", ""
    )

    pulumi_root = get_pulumi_root()
    os.makedirs(pulumi_root, exist_ok=True)
    os.environ["PULUMI_BACKEND_URL"] = os.environ.get(
        "PULUMI_BACKEND_URL", f"file://{pulumi_root}"
    )



import pulumi
import pulumi_aws as aws
import pulumi_eks as eks
import pulumi_kubernetes as k8s
import pulumi_kubernetes.helm.v3 as helm

from paka.cluster.aws.utils import odic_role_for_sa
from paka.config import CloudConfig
from paka.utils import call_once


@call_once
def create_cluster_autoscaler(
    config: CloudConfig,
    cluster: eks.Cluster,
    k8s_provider: k8s.Provider,
) -> None:
    """
    Sets up the cluster autoscaler for an EKS cluster.

    Args:
        cluster (eks.Cluster): The EKS cluster.
        k8s_provider (k8s.Provider): The Kubernetes provider.
        config (CloudConfig): The cluster config provided by user.

    Returns:
        None
    """
    project = config.cluster.name

    autoscaler_policy_doc = aws.iam.get_policy_document(
        statements=[
            aws.iam.GetPolicyDocumentStatementArgs(
                actions=[
                    "autoscaling:DescribeAutoScalingGroups",
                    "autoscaling:DescribeAutoScalingInstances",
                    "autoscaling:DescribeLaunchConfigurations",
                    "autoscaling:DescribeTags",
                    "autoscaling:SetDesiredCapacity",
                    "autoscaling:TerminateInstanceInAutoScalingGroup",
                    "ec2:DescribeLaunchTemplateVersions",
                    "eks:DescribeNodegroup",
                ],
                resources=["*"],
            )
        ]
    )

    autoscaler_policy = aws.iam.Policy(
        f"{project}-autoscaler-policy", policy=autoscaler_policy_doc.json
    )

    # The OIDC provider is required because the cluster autoscaler runs within the Kubernetes
    # cluster and needs to interact with the AWS API to manage the Auto Scaling Groups (ASGs).
    # OIDC provides a secure mechanism for the cluster autoscaler to authenticate with the AWS API.
    autoscaler_role = odic_role_for_sa(
        config, cluster, "autoscaler", "kube-system:cluster-autoscaler"
    )

    aws.iam.RolePolicyAttachment(
        f"{project}-autoscaler-role-policy-attachment",
        policy_arn=autoscaler_policy.arn,
        role=autoscaler_role.name,
    )

    helm.Chart(
        "cluster-autoscaler",
        helm.ChartOpts(
            chart="cluster-autoscaler",
            version="9.34.0",
            namespace="kube-system",
            fetch_opts=helm.FetchOpts(repo="https://kubernetes.github.io/autoscaler"),
            values={
                "autoDiscovery": {"clusterName": cluster.eks_cluster.name},
                "awsRegion": config.cluster.region,
                "rbac": {
                    "create": True,
                    "serviceAccount": {
                        "create": True,
                        "name": "cluster-autoscaler",
                        "annotations": {
                            "eks.amazonaws.com/role-arn": autoscaler_role.arn
                        },
                    },
                },
                "serviceMonitor": {"interval": "2s"},
                "image": {"tag": "v1.28.2"},
            },
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )



import contextlib
import os
import re
import select
import socket
import threading
import time
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    TypeAlias,
)

from kubernetes import client, config, watch
from kubernetes.client.rest import ApiException
from kubernetes.stream import portforward
from ruamel.yaml import YAML

from paka.logger import logger
from paka.utils import get_project_data_dir, read_yaml_file

KubernetesResourceKind: TypeAlias = Literal[
    "Deployment",
    "Service",
    "HorizontalPodAutoscaler",
    "ScaledObject",
    "TriggerAuthentication",
    "ServiceAccount",
    "Secret",
    "RoleBinding",
    "ConfigMap",
    "Role",
    "Package",
    "Environment",
    "Function",
    "Gateway",
    "VirtualService",
    "ServiceMonitor",
]


class CustomResource:
    def __init__(
        self,
        api_version: str,
        kind: KubernetesResourceKind,
        plural: str,
        spec: Dict[str, Any],
        metadata: client.V1ObjectMeta,
        status: Optional[Dict[str, Any]] = None,
    ):
        # Ensure api_version is in the format group/version
        if not re.match(r"^.+/v[\w]+$", api_version):
            raise ValueError("api_version must be in the format 'group/version'")
        self.api_version = api_version
        self.group, self.version = api_version.split("/")
        self.kind = kind
        self.plural = plural
        self.metadata = metadata
        self.spec = spec
        self.status = status


def create_namespaced_custom_object(namespace: str, resource: CustomResource) -> Any:
    body = {
        "apiVersion": resource.api_version,
        "kind": resource.kind,
        "metadata": {
            **resource.metadata.to_dict(),
            "name": resource.metadata.name,
            "namespace": namespace,
        },
        "spec": resource.spec,
    }

    if resource.status is not None:
        body["status"] = resource.status

    api_instance = client.CustomObjectsApi()

    return api_instance.create_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        body=body,
    )


def read_namespaced_custom_object(
    name: str, namespace: str, resource: CustomResource
) -> Any:
    api_instance = client.CustomObjectsApi()
    return api_instance.get_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        name=name,
    )


def replace_namespaced_custom_object(
    name: str, namespace: str, resource: CustomResource
) -> Any:
    res = read_namespaced_custom_object(name, namespace, resource)
    body = {
        "apiVersion": resource.api_version,
        "kind": resource.kind,
        "metadata": {
            **resource.metadata.to_dict(),
            "name": resource.metadata.name,
            "namespace": namespace,
            "resourceVersion": res["metadata"]["resourceVersion"],
        },
        "spec": resource.spec,
    }
    api_instance = client.CustomObjectsApi()
    return api_instance.replace_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        name=name,
        body=body,
    )


def delete_namespaced_custom_object(
    name: str, namespace: str, resource: CustomResource
) -> Any:
    api_instance = client.CustomObjectsApi()
    return api_instance.delete_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        name=name,
    )


def list_namespaced_custom_object(namespace: str, resource: CustomResource) -> Any:
    api_instance = client.CustomObjectsApi()
    custom_resources = api_instance.list_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
    )
    return custom_resources.get("items", [])


class KubernetesResource(Protocol):
    metadata: client.V1ObjectMeta
    kind: KubernetesResourceKind


def apply_resource(
    resource: KubernetesResource,
) -> Any:
    """
    Applies a Kubernetes resource by creating or updating it.

    Args:
        resource (KubernetesResource): The Kubernetes resource to apply.

    Returns:
        Any: The response from the API call.

    Raises:
        ValueError: If the resource kind is unsupported.
        ApiException: If an error occurs while creating or updating the resource.
    """

    # Determine the resource kind and prepare the appropriate API client
    kind = resource.kind
    namespace = resource.metadata.namespace
    if not namespace:
        raise ValueError("Namespace is required")

    if kind == "Deployment":
        api = client.AppsV1Api()
        create_method = api.create_namespaced_deployment
        replace_method = api.replace_namespaced_deployment
        read_method = api.read_namespaced_deployment
    elif kind == "Service":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_service
        replace_method = api.replace_namespaced_service
        read_method = api.read_namespaced_service
    elif kind == "HorizontalPodAutoscaler":
        api = client.AutoscalingV2Api()
        create_method = api.create_namespaced_horizontal_pod_autoscaler
        replace_method = api.replace_namespaced_horizontal_pod_autoscaler
        read_method = api.read_namespaced_horizontal_pod_autoscaler
    elif kind in [
        "ScaledObject",
        "TriggerAuthentication",
        "Package",
        "Environment",
        "Function",
        "Gateway",
        "VirtualService",
        "ServiceMonitor",
    ]:
        create_method = create_namespaced_custom_object
        replace_method = replace_namespaced_custom_object
        read_method = partial(read_namespaced_custom_object, resource=resource)
    elif kind == "ServiceAccount":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_service_account
        replace_method = api.patch_namespaced_service_account
        read_method = api.read_namespaced_service_account
    elif kind == "Secret":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_secret
        replace_method = api.patch_namespaced_secret
        read_method = api.read_namespaced_secret
    elif kind == "RoleBinding":
        api = client.RbacAuthorizationV1Api()
        create_method = api.create_namespaced_role_binding
        replace_method = api.patch_namespaced_role_binding
        read_method = api.read_namespaced_role_binding
    elif kind == "Role":
        api = client.RbacAuthorizationV1Api()
        create_method = api.create_namespaced_role
        replace_method = api.patch_namespaced_role
        read_method = api.read_namespaced_role
    elif kind == "ConfigMap":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_config_map
        replace_method = api.patch_namespaced_config_map
        read_method = api.read_namespaced_config_map
    else:
        raise ValueError(f"Unsupported kind: {kind}")

    # Try to read (get) the resource; if it exists, replace it, otherwise create it
    try:
        read_method(resource.metadata.name, namespace)
        response = replace_method(resource.metadata.name, namespace, resource)
        logger.info(f"{kind} '{resource.metadata.name}' updated.")
    except ApiException as e:
        if e.status == 404:
            response = create_method(namespace, resource)
            logger.info(f"{kind} '{resource.metadata.name}' created.")
        else:
            raise e
    return response


def create_namespace(name: str) -> None:
    """
    Creates a Kubernetes namespace.

    Args:
        name (str): The name of the namespace to create.

    Returns:
        None

    Raises:
        ApiException: If an error occurs while creating the namespace.
    """
    api = client.CoreV1Api()
    namespace = client.V1Namespace(metadata=client.V1ObjectMeta(name=name))
    try:
        api.create_namespace(body=namespace)
    except ApiException as e:
        if e.status == 409:  # Conflict, namespace already exists
            pass
        else:
            raise


def create_service_account(namespace: str, account_name: str) -> None:
    service_account = client.V1ServiceAccount(
        kind="ServiceAccount",
        metadata=client.V1ObjectMeta(name=account_name, namespace=namespace),
    )
    apply_resource(service_account)


def create_role_binding(
    binding_namespace: str,
    binding_name: str,
    role_name: str,
    subject_namespace: str,
    service_account_name: str,
) -> None:
    role_binding = client.V1RoleBinding(
        kind="RoleBinding",
        metadata=client.V1ObjectMeta(name=binding_name, namespace=binding_namespace),
        subjects=[
            client.V1Subject(
                kind="ServiceAccount",
                name=service_account_name,
                namespace=subject_namespace,
            )
        ],
        role_ref=client.V1RoleRef(
            api_group="rbac.authorization.k8s.io", kind="Role", name=role_name
        ),
    )
    apply_resource(role_binding)


def create_config_map(namespace: str, map_name: str, data: Dict[str, Any]) -> None:
    config_map = client.V1ConfigMap(
        kind="ConfigMap",
        metadata=client.V1ObjectMeta(name=map_name, namespace=namespace),
        data=data,
    )
    apply_resource(config_map)


def create_role(
    namespace: str, role_name: str, rules: List[client.V1PolicyRule]
) -> None:
    role = client.V1Role(
        api_version="rbac.authorization.k8s.io/v1",
        kind="Role",
        metadata=client.V1ObjectMeta(name=role_name, namespace=namespace),
        rules=rules,
    )
    apply_resource(role)


def is_ready_pod(pod: Any) -> bool:
    for condition in pod.status.conditions:
        if condition.type == "Ready" and condition.status == "True":
            return True
    return False


def find_free_port() -> int:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def run_port_forward(
    label_selector: str,
    local_port: int,
    container_port: int,
    namespace: Optional[str] = None,
) -> Tuple[threading.Event, Callable[[], None]]:
    v1 = client.CoreV1Api()

    if namespace is None:
        namespace = ""

    pods = v1.list_namespaced_pod(namespace, label_selector=label_selector)

    if len(pods.items) == 0:
        raise Exception(
            f"No available pod for port-forwarding with label selector {label_selector}"
        )

    namespaces: Dict[str, Any] = {}
    ns_list = []

    for pod in pods.items:
        if pod.metadata.namespace not in namespaces:
            namespaces[pod.metadata.namespace] = []
            ns_list.append(pod.metadata.namespace)
        namespaces[pod.metadata.namespace].append(pod)

    if len(ns_list) > 1:
        raise Exception(
            f"Found pods in {len(namespaces)} namespaces, {', '.join(ns_list)}. Please specify a namespace."
        )

    ns = ns_list[0]
    pods = namespaces.get(ns)

    if pods is None:
        raise Exception(f"Error finding pods within the given namespace {ns}")

    pod_name = None
    pod_namespace = None

    for pod in pods:
        if is_ready_pod(pod):
            pod_name = pod.metadata.name
            pod_namespace = pod.metadata.namespace
            break

    if pod_name is None or pod_namespace is None:
        raise Exception(
            f"No ready pod for port-forwarding with label selector {label_selector}"
        )

    ready_event = threading.Event()
    stop_event = threading.Event()

    def _proxy(
        client_socket: socket.socket,
        unix_socket: socket.socket,
        socket_lock: threading.Lock,
    ) -> None:
        try:
            while not stop_event.is_set():
                # Wait for data to be available on either socket
                readable, _, _ = select.select([client_socket, unix_socket], [], [])

                for sock in readable:
                    with socket_lock:
                        data = sock.recv(4096)

                    if not data:
                        # If the connection was closed, stop the proxy
                        return

                    # Determine the target socket
                    target_socket = (
                        unix_socket if sock is client_socket else client_socket
                    )

                    # Forward the data to the other socket
                    with socket_lock:
                        target_socket.sendall(data)

        except Exception as e:
            pass
        finally:
            client_socket.close()

    def _run_forward() -> None:
        pf = portforward(
            v1.connect_get_namespaced_pod_portforward,
            pod_name,
            pod_namespace,
            ports=str(container_port),
        )

        while not pf.connected:
            time.sleep(0.1)

        unix_socket = pf.socket(container_port)

        inet_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        inet_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        inet_socket.bind(("localhost", local_port))
        inet_socket.listen(5)
        inet_socket.settimeout(1)

        ready_event.set()

        lock = threading.Lock()

        try:
            while not stop_event.is_set():
                try:
                    # Accept new client connections
                    client_socket, addr = inet_socket.accept()
                    logger.debug(f"New connection from {addr}")

                    # Start a new thread to handle this client
                    threading.Thread(
                        target=_proxy, args=(client_socket, unix_socket, lock)
                    ).start()
                except socket.timeout:
                    # The accept call timed out, just loop again
                    continue
        finally:
            inet_socket.close()

        pf.close()

    forward_thread = threading.Thread(target=_run_forward)
    forward_thread.start()

    def stop_port_forward() -> None:
        stop_event.set()
        forward_thread.join()

    return ready_event, stop_port_forward


def setup_port_forward(
    label_selector: str, namespace: str, container_port: int
) -> Tuple[str, Callable[[], None]]:
    local_port = find_free_port()

    max_duration = 2

    ready_event, stop_forward = run_port_forward(
        label_selector, local_port, container_port, namespace
    )

    ready_event.wait()

    logger.debug(f"Waiting for port forward {local_port} to start...")

    wait_duration = 0.05
    while True:
        try:
            with socket.create_connection(
                ("localhost", local_port), timeout=wait_duration
            ):
                break
        except:
            logger.debug(f"Error dialing on local port {local_port}")
            time.sleep(wait_duration)
            wait_duration *= 2
            if wait_duration > max_duration:
                wait_duration = max_duration

    logger.debug(f"Port forward from local port {local_port} started")

    return str(local_port), stop_forward


def try_load_kubeconfig() -> bool:
    try:
        config.load_kube_config()
        return True
    except Exception as e:
        logger.debug(f"Error loading kubeconfig: {e}")
        return False


class KubeconfigMerger:
    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
        self.config = config or {}

    def _entries_by_key(self, key: str) -> List[Any]:
        self.config[key] = self.config.get(key) or []
        entries = self.config[key]
        if not isinstance(entries, list):
            raise Exception(
                f"Tried to insert into {key}, "
                f"which is a {type(entries)} "
                f"not a list."
            )
        return entries

    def _index_same_name(
        self, entries: List[Any], new_entry: Dict[str, Any]
    ) -> Optional[int]:
        if "name" in new_entry:
            name_to_search = new_entry["name"]
            for i, entry in enumerate(entries):
                if "name" in entry and entry["name"] == name_to_search:
                    return i
        return None

    def insert_entry(self, key: str, new_entry: Any) -> None:
        entries = self._entries_by_key(key)
        same_name_index = self._index_same_name(entries, new_entry)
        if same_name_index is None:
            entries.append(new_entry)
        else:
            entries[same_name_index] = new_entry

    def merge(self, new_config: Dict[str, Any]) -> None:
        for cluster in new_config.get("clusters", []):
            self.insert_entry("clusters", cluster)
        for user in new_config.get("users", []):
            self.insert_entry("users", user)
        for context in new_config.get("contexts", []):
            self.insert_entry("contexts", context)

        self.config["current-context"] = new_config["current-context"]

        for key in new_config.keys():
            if key not in ["clusters", "users", "contexts", "current-context"]:
                self.config[key] = new_config[key]


def update_kubeconfig() -> None:
    system_kubeconfig_path = os.path.expanduser("~/.kube/config")
    current_config = read_yaml_file(system_kubeconfig_path)
    merger = KubeconfigMerger(current_config)

    cluster_kubeconfig_path = os.path.join(
        get_project_data_dir(), "current_cluster", "kubeconfig.yaml"
    )
    new_config = read_yaml_file(cluster_kubeconfig_path)
    merger.merge(new_config)

    # Convert OrderedDict to dict and sort keys
    sorted_config = {k: merger.config[k] for k in sorted(merger.config)}

    # Dump the sorted config to a YAML-formatted string
    with open(system_kubeconfig_path, "w") as file:
        yaml = YAML()
        yaml.dump(sorted_config, file)


def tail_logs(namespace: str, pod_name: str, container_name: str) -> None:
    v1 = client.CoreV1Api()
    w = watch.Watch()

    while True:
        pod = v1.read_namespaced_pod(name=pod_name, namespace=namespace)
        if pod.status.phase == "Running":
            break
        elif pod.status.phase in ["Failed", "Succeeded"]:
            logger.info(f"\nPod {pod_name} is in phase {pod.status.phase}")
            return
        else:
            # print dot on the same line
            print(".", end="", flush=True)
            time.sleep(1)

    print("\n")
    for event in w.stream(
        v1.read_namespaced_pod_log,
        namespace=namespace,
        name=pod_name,
        container=container_name,
    ):
        logger.info(event)


def remove_crd_finalizers(name: str) -> None:
    api = client.ApiextensionsV1Api()
    body = [{"op": "remove", "path": "/metadata/finalizers"}]
    api.patch_custom_resource_definition(name, body)


# Load the kubeconfig when this module is imported
try_load_kubeconfig()



import pulumi_aws as aws
import pulumi_kubernetes as k8s

from paka.cluster.fluentbit import create_fluentbit
from paka.config import CloudConfig
from paka.constants import PROJECT_NAME

LOG_GROUP = f"EKSContainerLogs/{PROJECT_NAME}"


def enable_cloudwatch(cloud_config: CloudConfig, k8s_provider: k8s.Provider) -> None:
    aws.cloudwatch.LogGroup(
        "log-group",
        name=LOG_GROUP,
        retention_in_days=cloud_config.cluster.logRetentionDays,
    )

    # Fluent Bit configuration for forwarding logs to CloudWatch
    fluent_bit_config = f"""
[SERVICE]
    Parsers_File /fluent-bit/etc/parsers.conf

[INPUT]
    Name              tail
    Path              /var/log/containers/*.log
    Parser            docker
    Tag               kube.*
    Refresh_Interval  5

[OUTPUT]
    Name               cloudwatch_logs
    Match              kube.*
    log_group_name     {LOG_GROUP}
    log_stream_prefix  eks/
    region             {cloud_config.cluster.region}
"""
    create_fluentbit(fluent_bit_config, k8s_provider)



import shlex
import subprocess

from paka.logger import logger
from paka.utils import random_str


def push_to_ecr(
    image_name: str, repository_uri: str, aws_region: str, app_name: str
) -> str:
    """
    Pushes a Docker image to an Amazon ECR repository.

    This function tags the Docker image with a version tag and the "latest" tag,
    logs in to the ECR repository, and pushes the image to the repository.
    The version tag is generated randomly.

    All applications share the same container registry repository.
    To differentiate between them, we append the application name to the image tag.
    The '-latest' suffix is added to handle cases where applications themselves are tagged.
    This ensures that even tagged applications have a unique identifier in the shared repository.

    Args:
        image_name (str): The name of the Docker image to push.
        repository_uri (str): The URI of the ECR repository to push the image to.
        aws_region (str): The AWS region where the ECR repository is located.
        app_name (str): The name of the application. Used to generate the image tags.

    Raises:
        subprocess.CalledProcessError: If an error occurs while executing a subprocess command.

    Returns:
        str: The version tag of the image that was pushed.
    """
    try:
        # Get ECR login password
        login_password = (
            subprocess.check_output(
                ["aws", "ecr", "get-login-password", "--region", aws_region]
            )
            .decode()
            .strip()
        )

        # Generate a random version number
        version = random_str()

        # Tag the image with the repository URI and the version tag
        version_tag = f"{app_name}-v{version}"

        # Tag the image with the repository URI
        subprocess.run(
            [
                "docker",
                "tag",
                f"{image_name}:latest",
                f"{repository_uri}:{version_tag}",
            ],
            check=True,
        )

        # Tag the image with the repository URI and the "latest" tag
        latest_tag = f"{app_name}-latest"
        subprocess.run(
            [
                "docker",
                "tag",
                f"{image_name}:latest",
                f"{repository_uri}:{latest_tag}",
            ],
            check=True,
        )

        # Perform Docker login and push in one command
        login_push_cmd = (
            f"echo {shlex.quote(login_password)} | "
            f"docker login --username AWS --password-stdin {repository_uri} && "
            f"docker push {repository_uri}:{version_tag} && "
            f"docker push {repository_uri}:{latest_tag}"
        )
        subprocess.run(login_push_cmd, shell=True, check=True, executable="/bin/sh")

        logger.info(f"Successfully pushed {image_name} to {repository_uri}")
        return version_tag
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred: {e}")
        raise



from collections import namedtuple

Model = namedtuple("Model", ["name", "url", "sha256"])

SUPPORTED_MODELS = {
    "llama2-7b": Model(
        name="llama2-7b",
        url="https://huggingface.co/TheBloke/Llama-2-7B-Chat-GGUF/resolve/main/llama-2-7b-chat.Q4_0.gguf",
        sha256="9958ee9b670594147b750bbc7d0540b928fa12dcc5dd4c58cc56ed2eb85e371b",
    ),
    "mistral-7b": Model(
        name="mistral-7b",
        url="https://huggingface.co/TheBloke/Mistral-7B-Instruct-v0.2-GGUF/resolve/main/mistral-7b-instruct-v0.2.Q4_0.gguf",
        sha256="25d80b918e4432661726ef408b248005bebefe3f8e1ac722d55d0c5dcf2893e0",
    ),
    "codellama-7b": Model(
        name="codellama-7b",
        url="https://huggingface.co/TheBloke/CodeLlama-7B-GGUF/resolve/main/codellama-7b.Q4_0.gguf",
        sha256="33052f6dd41436db2f83bd48017b6fff8ce0184e15a8a227368b4230f1da97b5",
    ),
    "gte-base": Model(
        name="gte-base",
        url="https://huggingface.co/jjleng/gte-base-gguf/resolve/main/gte-base.q4_0.gguf",
        sha256="2413866ece3b8b9eedf6c2a4393d4b56dbfa363c173ca3ba3a2f2a44db158982",
    ),
}
